#!/usr/bin/env python3

import pandas as pd
import numpy as np

from configparser import ConfigParser

from domaintools import API
from domaintools.exceptions import BadRequestException, NotFoundException

from tqdm.auto import tqdm

import os.path

from functools import reduce

from .decorators import retry
from .data import flatten as huntlib_data_flatten

__all__ = ['DomainTools']

class DomainTools(object):
    '''
    The DomainTools class allows you to easily perform some common types of calls
    to the DomainTools API.  It uses their official `domaintools_api` Python module
    to do most of the work but is not a complete replacement for that module. In
    particular, this class concentrates on a few calls that are most relevant for
    data analytic style threat hunting (risk & reputation scores, WHOIS info, etc).

    This most methods pass through any kwargs to the underlying domaintools methods, 
    with one important exception: the class consults the user's ~/.huntlibrc 
    (if present) to determine the API username and key so you don't always have to provide
    them during authentication.

    :param api_username: The API authentication username (OPTIONAL)
    :type api_username: str
    :param api_key: The API authentication key (OPTIONAL)
    :type api_key: str
    :param `**kwargs`: Additional keyword args are passed to the underlying domaintools module init
    '''

    _DEFAULT_CONFIG_FILE = os.path.expanduser("~/.huntlibrc")

    _handle = None # API Handle
    _account_information = None # Cached information about the API user's account
    _available_api_calls = None # Cached list of API endpoints the account has access to

    def __init__(self, *args, **kwargs):

        # Read our config file, if it exists
        config = ConfigParser()
        config.read(self._DEFAULT_CONFIG_FILE)
        
        api_username = ""
        api_key = ""

        if 'api_username' in kwargs:
            api_username = kwargs['api_username']
        elif config.has_option('domaintools', 'api_username'):
            api_username = config.get('domaintools', 'api_username')

        if 'api_key' in kwargs:
            api_key = kwargs['api_key']
        elif config.has_option('domaintools', 'api_key'):
            api_key = config.get('domaintools', 'api_key')

        # Remove these from kwargs, so we don't have duplicate args.
        kwargs.pop('api_username', None)
        kwargs.pop('api_key', None)

        self.authenticate(
            api_username=api_username,
            api_key=api_key,
            **kwargs
        )

        self._account_information = self.account_information(force_refresh=True)
        self._available_api_calls = self.available_api_calls(force_refresh=True)

    @retry()
    def authenticate(self, api_username="", api_key="", **kwargs):
        """
        Authenticate to the DomainTools API. Calling this function directly is OK,
        but you won't get the benefit of consulting ~/.huntlibrc for default creds
        if you do.  

        :param api_username: The API authentication username (OPTIONAL)
        :type api_username: str
        :param api_key: The API authentication key (OPTIONAL)
        :type api_key: str
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Exceptions:
          Will raise ValueError if the api_username or api_key are not provided
          or are of the wrong type.
        """
        # Sanity check the provided authentication info
        if not api_username:
            raise ValueError("You must supply a value for 'api_username'.")
        elif not isinstance(api_username, str):
            raise ValueError("The 'api_username' field must be a string.")

        if not api_key:
            raise ValueError("You must supply a value for 'api_key'.")
        elif not isinstance(api_key, str):
            raise ValueError("The 'api_key' field must be a string.")


        # Actually authenticate now
        self._handle = API(
            api_username,
            api_key,
            **kwargs
        )

    @retry()
    def account_information(self, force_refresh=False, **kwargs):
        '''
        Return a dict containing information about limits and usage of the various
        domaintools API calls for the authenticated API user.

        :param force_refresh: A boolean controlling whether or not to refresh the cached info
        :type force_refresh: bool
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value:
        A single dict, where each key is the name of an endpoint from the underlying
        domaintools API, and the values are dicts containing detailed about that endpoint.

        For example:

        {
            'domain-profile': {
                'per_month_limit': None, 
                'per_minute_limit': '180', 
                'absolute_limit': None, 
                'usage': {
                    'today': '0', 
                    'month': '100'
                }, 
                'expiration_date': '2020-12-31'
            }, 
            'whois': {
                'per_month_limit': None, 
                'per_minute_limit': '180', 
                'absolute_limit': None, 
                'usage': {
                    'today': '9', 
                    'month': '997'
                }, 
                'expiration_date': '2020-12-31'
            }
        }
        '''

        if self._account_information and not force_refresh:
            return self._account_information

        res = self._handle.account_information(**kwargs)

        if res:
            info = dict()
            for item in res:
                id = item.pop('id')
                info[id] = dict(item)
            self._account_information = info

        return self._account_information

    @retry()
    def available_api_calls(self, force_refresh=False, **kwargs):
        '''
        Returns a list of endpoints available to the authenticated API user.

        :param force_refresh: A boolean controlling whether or not to refresh the cached info
        :type force_refresh: bool
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value:

        A list of strings containing the API endpoint names.

        For example:

        [
            'domain_profile', 
            'whois', 
            'whois_history', 
            'reverse_ip', 
            ...
        ]
        '''

        if self._available_api_calls and not force_refresh:
            return self._available_api_calls

        res = self._handle.available_api_calls(**kwargs)

        if res:
            self._available_api_calls = list(res)

        return self._available_api_calls

    @retry()
    def whois(self, query=None, **kwargs):
        '''
        Return basic WHOIS info for a given domain or IP address.

        :param query: A domain or IP address
        :param type: string
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value:
        A dict containing basic WHOIS information.

        For example:

        {
            'registrant': 'Google LLC', 
            'registration': {
                'created': '1997-09-15', 
                'expires': '2028-09-14', 
                'updated': '2019-09-09', 
                'registrar': 'MarkMonitor Inc.', 
                'statuses': [
                    'clientDeleteProhibited', 
                    'clientTransferProhibited', 
                    'clientUpdateProhibited', 
                    'serverDeleteProhibited', 
                    'serverTransferProhibited', 
                    'serverUpdateProhibited'
                ]
            }, 
            'name_servers': [
                'NS1.GOOGLE.COM', 
                'NS2.GOOGLE.COM', 
                'NS3.GOOGLE.COM', 
                'NS4.GOOGLE.COM'
            ], 
            'whois': {
                'date': '2020-07-12', 
                'record': 'Domain Name: google.com\nRegistry Domain ID: 2138514_DOMAIN_COM-VRSN\n...'
            }, 
            'record_source': 'google.com'
        }

        :Exceptions:
        Will raise ValueError if no query is supplied, or if the query is not a string.

        '''

        if not query:
            raise ValueError("You must supply either a domain or an IP address.")
        elif not isinstance(query, str):
            raise ValueError("The query parameter must be a string.")
        
        try:
            whois_info = dict(
                list(
                    self._handle.whois(query, **kwargs)
                )
            )
        except (BadRequestException, NotFoundException):
            return dict()

        return whois_info

    @retry()
    def parsed_whois(self, query=None, flatten=False, **kwargs):

        '''
        Return extended WHOIS info for a given domain or IP address.

        :param query: A domain or IP address
        :param type: string
        :param flatten: A boolean controlling whether to attempt to normalize the nested dicts or lists into a single flat dict (DEFAULT False)
        :type flatten: bool 
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value:
        A dict containing basic and extended WHOIS information.

        For example:

        {
            ...
            'parsed_whois': {
                'domain': 'google.com', 
                'created_date': '1997-09-15T00:00:00-07:00', 
                'updated_date': '2019-09-09T08:39:04-07:00', 
                'expired_date': '2028-09-13T00:00:00-07:00', 
                'statuses': [
                    'clientDeleteProhibited', 
                    'clientTransferProhibited', 
                    'clientUpdateProhibited', 
                    'serverDeleteProhibited', 
                    'serverTransferProhibited', 
                    'serverUpdateProhibited'
                ], 
                'name_servers': [
                    'ns1.google.com', 
                    'ns2.google.com', 
                    'ns3.google.com', 
                    'ns4.google.com'
                ], 
                'registrar': {
                    'name': 'MarkMonitor, Inc. MarkMonitor Inc.', 
                    'abuse_contact_phone': '12083895770', 
                    'abuse_contact_email': 'abusecomplaints@markmonitor.com', 
                    'iana_id': '292', 
                    'url': 'http://www.markmonitor.com', 
                    'whois_server': 'whois.markmonitor.com'
                }, 
                'contacts': {
                    'registrant': {
                        'name': '', 
                        'org': 'Google LLC', 
                        'street': [], 
                        'city': '', 
                        'state': 'CA', 
                        'postal': '', 
                        'country': 'us', 
                        'phone': '', 
                        'fax': '', 
                        'email': 'REDACTED FOR PRIVACY (DT)'
                    }
                    ...
                }
            }
            ...
        }

        :Exceptions:
        Will raise ValueError if no query is supplied, or if the query is not a string.

        '''

        if not query:
            raise ValueError(
                "You must supply either a domain or an IP address.")
        elif not isinstance(query, str):
            raise ValueError(
                "The query parameter must be a string.")

        try:
            whois_info = dict(
                list(
                    self._handle.parsed_whois(query, **kwargs)
                )
            )
        except (BadRequestException, NotFoundException):
            return dict()

        
        if flatten:
            # Normalize the nested dictionary keys into a single level.
            whois_info = huntlib_data_flatten(whois_info)

        return whois_info

    @retry()
    def brand_monitor(self, query=None, **kwargs):
        '''
        Given a query string containing one or more search terms (separated by '|'),
        return a list of any newly-active or pending domain registrationations 
        containing ALL of the terms.

        :param query: A string containing one or more search terms (separated by '|')
        :type query: string
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value:

        A list of dicts, with each dict containing a result.

        For example:

        [
            {
                'domain': '54google.com', 
                'status': 'new'
            }, 
            {
                'domain': 'aboutmicrosoftandgoogleapps.com',
                 'status': 'on-hold'
            },
            ...
        ]

        :Exceptions:

        Will raise ValueError if no query is supplied or if the query is not a string.
        '''

        if not query:
            raise ValueError("You must specify a query pattern.")
        elif isinstance(query, list):
            query = "|".join(query)
        elif not isinstance(query, str):
            raise ValueError("The 'query' parameter must be either a string or a list of strings.")

        return list(self._handle.brand_monitor(query, **kwargs))

    @retry()
    def domain_profile(self, query=None, flatten=False, **kwargs):
        '''
        Look up basic information about a domain, including DNS, WHOIS, history and
        web site info along with pointers to more detailed info.

        :param query: The domain to look up
        :type query: string
        :param flatten: A boolean controlling whether to attempt to normalize the nested dicts/lists into a single flat dict (DEFAULT False)
        :type flatten: bool 
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value:

        A dict containing the various pieces of info.  

        For example:

        {
            'registrant': {
                'name': 'Google LLC', 
                'domains': 18696, 
                'product_url': 'https://reversewhois.domaintools.com/?all[]=Google+LLC&none[]='
            }, 
            'server': {
                'ip_address': '172.217.14.196', 
                'other_domains': 151, 
                'product_url': 'https://reverseip.domaintools.com/search/?q=google.com'
            }, 
            'registration': {
                'created': '1997-09-15', 
                'expires': '2028-09-14', 
                'updated': '2019-09-09', 
                'registrar': 'MarkMonitor Inc.', 
                'statuses': [
                    'clientDeleteProhibited', 
                    'clientTransferProhibited', 
                    'clientUpdateProhibited', 
                    'serverDeleteProhibited', 
                    'serverTransferProhibited', 
                    'serverUpdateProhibited'
                ]
            }, 
            'name_servers': [
                {
                    'server': 'NS1.GOOGLE.COM', 
                    'product_url': 'https://reversens.domaintools.com/search/?q=NS1.GOOGLE.COM'
                }, 
                {
                    'server': 'NS2.GOOGLE.COM', 
                    'product_url': 'https://reversens.domaintools.com/search/?q=NS2.GOOGLE.COM'
                }, 
                {
                    'server': 'NS3.GOOGLE.COM', 
                    'product_url': 'https://reversens.domaintools.com/search/?q=NS3.GOOGLE.COM'
                }, 
                {
                    'server': 'NS4.GOOGLE.COM', 
                    'product_url': 'https://reversens.domaintools.com/search/?q=NS4.GOOGLE.COM'
                }
            ], 
            ...
        }

        :Exceptions:

        Raises ValueError if no query is supplied or if it is not a string.
        '''
        if not query:
            raise ValueError("You must specify a query domain.")
        elif not isinstance(query, str):
            raise ValueError("query parameter must be a string.")

        profile = dict(
            list(
                self._handle.domain_profile(query, **kwargs)
            )
        )

        if flatten:
            # Normalize the nested dictionary keys into a single level.
            profile = huntlib_data_flatten(profile)

        return profile

    @retry()
    def domain_reputation(self, domain=None, reasons=False, **kwargs):
        '''
        Return a risk score based on the reputation of the given domain, 
        with an optional list of reasons contributing to the score.

        :param domain: The domain for which to retrieve the score
        :type domain: string
        :param reasons: Determines whether or not to include a list of reasons for the score (DEFAULT False)
        :type reasons: bool
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value: 

        A Dict containing the requested information.  If the domain cannot be found, return
        an empty dict.

        For example:

        {
            'domain': 'domaintools.xyz', 
            'risk_score': 18.69, 
            'reasons': [
                'registrant'
            ]
        }

        :Exceptions:

        Raises ValueError if no domain is provided, or if the domain is not a string.

        '''
        if not domain:
            raise ValueError("You must specify a query domain.")
        elif not isinstance(domain, str):
            raise ValueError("The domain parameter must be a string.")

        try:
            reputation = dict(
                list(
                    self._handle.reputation(domain, reasons, **kwargs)
                )
            )
        except (BadRequestException, NotFoundException):
            return dict()

        return reputation

    @retry()
    def risk(self, domain=None,  **kwargs):
        '''
        Return risk scores for a domain with respect to individual risk factors.

        :param domain: The domain for which to retrieve the score
        :type domain: string
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value: 

        A Dict containing the requested information.  If the domain cannot be found, return
        an empty dict.

        For example:

        {
            'proximity': 18, 
            'threat_profile': 36, 
            'threat_profile_phishing': 36, 
            'threat_profile_malware': 17, 
            'threat_profile_spam': 2
        }

        :Exceptions:

        Raises ValueError if no domain is provided, or if the domain is not a string.

        '''
        if not domain:
            raise ValueError("You must specify a query domain.")
        elif not isinstance(domain, str):
            raise ValueError("The domain parameter must be a string.")

        try:
            risk = self._handle.risk(domain, **kwargs)
            # Turn the list of individual dictionaries (with duplicate
            # keys) into a single dictionary.
            risk = {x['name']: x['risk_score'] for x in risk}
        except (BadRequestException, NotFoundException):
            return dict()


        return risk
    
    @retry() 
    def iris_enrich(self, query=None, flatten=False, asframe=False, **kwargs):
        '''
        Bulk enrichment for lists of domains against the DomainTools IRIS database.
        This will do basic deduplication (e.g., 'google.com' will only be looked up 
        once no matter how many times it appears in the input list, but 'google.com',
        'www.google.com' and 'drive.google.com' are not considered duplicates).  

        :param query: The domain(s) to enrich
        :type query: list or pandas Series object
        :param flatten: A boolean controlling whether to attempt to normalize the nested dicts/lists into a single flat dict (DEFAULT False)
        :type flatten: bool 
        :param asframe: Return the enriched data as a pandas DataFrame instead of a dict (DEFAULT False)
        :param `**kwargs`: Additional arguments to pass to the underlying domaintools module

        :Return Value:
        
        Returns a dict where each key is an enriched domain and the corresponding
        value is a dict with the enrichment data for that domain.  For example:

        {
            'google.com': {
                'whois_url': 'https://whois.domaintools.com/google.com',
                'active': True,
                [...]
            },
            'microsoft.com': {
                'whois_url': 'https://whois.domaintools.com/microsoft.com',
                'active': True,
                [...]
            }
        }

        If `asframe` is True, the result is returned instead as a pandas DataFrame 
        object, where the 'domain' column contains the enriched domains, with their
        enrichment data flattened into columns, like so:

                domain          whois_url                                   active  [...]
            0   google.com      https://whois.domaintools.com/google.com    True    [...]
            1   microsoft.com   https://whois.domaintools.com/microsoft.com True    [...]
        
        '''
        if query is None:
            raise ValueError("You must specify a domain or list of domains to query.")

        if isinstance(query, list) or isinstance(query, pd.core.series.Series):
            # Convert a list of strings to a single comma-separated string
            query = ','.join(query)
        elif not isinstance(query, str):
            raise ValueError("The query must be either a string or a list of strings.")
        
        try:
            enrich = list(self._handle.iris_enrich(query, **kwargs))
        except (BadRequestException, NotFoundException):
            return dict()

        data = dict()
        for i in enrich:
            if 'domain' in i:
                domain = i.pop('domain')
                if flatten:
                    data[domain] = huntlib_data_flatten(i)
                else:
                    data[domain] = i

        if asframe:
            return pd.DataFrame(data).transpose().reset_index().rename(columns={'index': 'domain'})
        else:
            return data 
    
    def enrich(self, df=None, column=None, prefix='dt_enrich.', progress_bar=False, fields=None, batch_size=100):
        '''
        Enrich a pandas DataFrame object with information from DomainTools.  Note that the 
        original DataFrame is not modified, so you must assign the return value to a variable
        if you want to keep it.  e.g. `df = dt.enrich(df, column='domains')`.

        :param df: The DataFrame to enrich
        :type df: pandas.DataFrame
        :param column: The name of the column containin domains and/or IPs to enrich (as strings)
        :type column: string
        :param prefix: Naming prefix for the newly-added columns (DEFAULT 'dt_whois.')
        :type prefix: string
        :param progress_bar: If True, attempt to show enrichment progress (DEFAULT False)
        :type progress_bar: bool
        :param fields: A list of specific enrichment field names to add (DEFAULT add all fields)
        :type fields: list of strings
        :param batch_size: The number of domains/IPs to enrich in one "batch" (DEFAULT 100)
        :type batch_size: integer

        :Return Value:

        A pandas DataFrame object containing all of the original information plus many 
        additional enrichment columns.

        :Exceptions:
        Raises ValueError if the required options are not present or are of the wrong type.

        '''

        if df is None:
            raise ValueError(
                "You must supply a pandas DataFrame in the 'df' parameter.")
        elif not isinstance(df, pd.core.frame.DataFrame):
            raise ValueError(
                "The argument for the 'df' parameter must be a pandas DataFrame.")

        if not column:
            raise ValueError("You must supply a column name to enrich.")
        elif not isinstance(column, str):
            raise ValueError("The column name must be a 'str'.")
        elif not column in df.columns:
            raise ValueError(
                f"The column '{column}' does not exist in the frame.")

        if prefix is None or not isinstance(prefix, str):
            raise ValueError("The column name prefix must be a 'str'.")

        if fields is not None and not isinstance(fields, list):
            raise ValueError(
                "The 'fields' parameter must be a list of strings.")

        # Attempt some basic deduplication to save API calls
        unique_domains = pd.Series(
            df[column].unique(),
            dtype='object'
        )

        if progress_bar:
            tqdm.pandas(desc='Enriching')

        enrichment_df = pd.DataFrame()

        with tqdm(desc="Enriching", total=unique_domains.size, disable=not progress_bar) as pbar:

            for batch in [unique_domains[i:i+batch_size] for i in range(0, unique_domains.size, batch_size)]:

                res = self.iris_enrich(
                    batch,
                    flatten=True,
                    asframe=True
                )

                # We have to do this the hard way, instead of just DataFrame(res)
                # because some of the items in res contain lists of unequal length,
                # which causes pandas to throw an exception.
                results = pd.DataFrame(
                    dict(
                        [(k, pd.Series(v, dtype='object')) for k,v in res.items()]
                    )
                )
                
                enrichment_df = enrichment_df.append(
                    results,
                    ignore_index=True
                )
            
                pbar.update(batch.size)

        enrichment_df = enrichment_df.add_prefix(prefix)

        # If we asked for only certain fields, filter for those
        if fields:
            if not f'{prefix}domain' in fields:
                # Make sure this is in the final fields list no matter what,
                # because we rely on it as a merge column below
                fields.append(f'{prefix}domain')
            enrichment_df = enrichment_df[fields]

        df = pd.merge(
            df,
            enrichment_df,
            how='left',
            left_on=column,
            right_on=f'{prefix}domain'
        )

        df = df.drop(f'{prefix}domain', axis='columns')

        return df



from huntlib.exceptions import AuthenticationErrorSearchException, InvalidRequestSearchException, UnknownSearchException
from builtins import object
from elasticsearch import Elasticsearch
from elasticsearch_dsl import Search
import elasticsearch.exceptions
import pandas as pd
from datetime import datetime, timedelta

class ElasticDF(object):
    '''
    The ElasticDF() class searches Elastic and returns results as a Pandas
    DataFrame.  This makes it easier to work with the search results with
    standard data analysis techniques.

    Example usage:

        # Create a plaintext connection to the Elastic server, no authentication
        e = ElasticDF(url="http://localhost:9200")

        # The same, but with SSL and authentication
        e = ElasticDF(url="https://localhost:9200", ssl=True, username="myuser",
                      password="mypass")

        # Fetch search results from an index or index pattern for the previous day
        df = e.search_df(lucene="item:5282 AND color:red", index="myindex-*", days=1)

        # The same, but do not flatten structures into individual columns.
        # This will result in each structure having a single column with a
        # JSON string describing the structure.
        df = e.search_df(lucene="item:5282 AND color:red", index="myindex-*", days=1,
                         normalize=False)

        # A more complex example, showing how to set the Elastic document type,
        # use Python-style datetime objects to constrain the search to a certain
        # time period, and a user-defined field against which to do the time
        # comparisons.
        df = e.search_df(lucene="item:5285 AND color:red", index="myindex-*",
                         doctype="doc", date_field="mydate",
                         start_time=datetime.now() - timedelta(days=8),
                         end_time=datetime.now() - timedelta(days=6))

    The search() and search_df() methods will raise
    InvalidRequestSearchException() in the event that the search request is
    syntactically correct but is otherwise invalid. For example, if you request
    more results be returned than the server is able to provide.
    They will raise AuthenticationErrorSearchException() in the event the server 
    denied the credentials during login.  They can also raise an
    UnknownSearchException() for other situations, in which case the exception
    message will contain the original error message returned by Elastic so you
    can figure out what went wrong.
    '''

    es_conn = None # The connection to the ES server

    def __init__(self, url=None, timeout=250, ssl=False, username="", password="", verify_certs=True, ca_certs=None):
        '''
        Create the ElasticDF object and log into the Elastic server.
        '''

        self.es_conn = Elasticsearch(
            url,
            timeout=timeout,
            use_ssl=ssl,
            verify_certs=verify_certs,
            ca_certs=ca_certs,
            http_auth=(username, password)
        )

    def search(self, lucene, index="*", doctype="doc", fields=None,
               date_field="@timestamp", days=None, start_time=None,
               end_time=None, limit=None):
        '''
        Search Elastic and return the results as a list of dicts.

        lucene: A string containing the Elastic search (e.g., 'item:5282 AND color:red')
        index: A string containing the index name to search, or an index name pattern
               if you want to search multiple indices (e.g., 'myindex' or 'myindex-*')
        doctype: The document type you are interested in.
        fields: A string containing a comma-separated list of field names to return.
                The default is to return all fields, but using this list you can
                select only certain fields, which may make things a bit faster.
        date_field: The name of the field used for date/time comparison.
        days: Search the past X days. If provided, this supercedes both start_time
              and end_time.
        start_time: A datetime() object representing the start of the search
                    window. If used without end_time, the end of the search
                    window is the current time.
        end_time: A datetime() object representing the end of the search window.
                  If used without start_time, the search start will be the earliest
                  time in the index.
        limit: An integer describing the max number of search results to return.
        '''

        s = Search(using=self.es_conn, index=index, doc_type=doctype)

        s = s.query("query_string", query=lucene)

        if fields:
            s = s.source(fields.split(','))

        # Add timestamp filters, if provided.  Days takes precendence over
        # use of either/both of start_time and end_time.
        # Note the weird unpacked dictionary syntax in the call to s.filter().
        # We have to do it this way because Python has an issue naming things
        # with "@" in them, but the default timestamp field in many ES servers is
        # "@timestamp".
        # ref:  https://github.com/elastic/elasticsearch-dsl-py/blob/master/docs/search_dsl.rst
        if days:
            end = datetime.now()
            start = end - timedelta(days=days)
            s = s.filter('range', ** {date_field: {"gte": start, "lte": end}})
        elif start_time and not end_time:
            s = s.filter('range', ** {date_field: {"gte": start_time}})
        elif end_time and not start_time:
            s = s.filter('range', ** {date_field: {"lte": end_time}})
        elif start_time and end_time:
            s = s.filter('range', ** {date_field: {"gte": start_time, "lte": end_time}})

        # Add a search limit, if one is specified. Note that this is per-shard,
        # not total results.  Since this is where the search actually runs (the
        # call to excute() does this) then we also have to handle authentication
        # issues.
        try:
            if limit:
                s = s.params(size=limit)
                response = s.execute()
            else:
                # Scan to explicitly return all results
                response = s.execute()
                s = s.scan()
        except elasticsearch.exceptions.AuthenticationException:
            raise AuthenticationErrorSearchException("Login failed.")

        if response.success():
            for hit in s:
                yield hit.to_dict()
        else:
            reason = response._shards.failures[0].reason
            if "Result window is too large" in reason['reason']:
                raise InvalidRequestSearchException("Too many results requested (more than index.max_result_window settings on the index). Either reduce the result limit or remove the limit paramter entirely.")
            else:
                raise UnknownSearchException("Message from Elastic was: %s" % reason['reason'])

    def search_df(self, lucene, index="*", doctype="doc", fields=None,
                  date_field="@timestamp", days=None, start_time=None,
                  end_time=None, normalize=True, limit=None):
        '''
        Search Elastic and return the results as a Pandas DataFrame.

        lucene: A string containing the Elastic search (e.g., 'item:5282 AND color:red')
        index: A string containing the index name to search, or an index name pattern
               if you want to search multiple indices (e.g., 'myindex' or 'myindex-*')
        doctype: The document type you are interested in.
        fields: A string containing a comma-separated list of field names to return.
                The default is to return all fields, but using this list you can
                select only certain fields, which may make things a bit faster.
        date_field: The name of the field used for date/time comparison.
        days: Search the past X days. If provided, this supercedes both start_time
              and end_time.
        start_time: A datetime() object representing the start of the search
                    window. If used without end_time, the end of the search
                    window is the current time.
        end_time: A datetime() object representing the end of the search window.
                  If used without start_time, the search start will be the earliest
                  time in the index.
        normalize: If set to True, fields containing structures (i.e. subfields)
                   will be flattened such that each field has it's own column in
                   the dataframe. If False, there will be a single column for the
                   structure, with a JSON string encoding all the contents.
        limit: An integer describing the max number of search results to return.
        '''
        results = list()

        for hit in self.search(lucene=lucene, index=index, doctype=doctype,
                               fields=fields, date_field=date_field, days=days,
                               start_time=start_time, end_time=end_time,
                               limit=limit):
            results.append(hit)

        if normalize:
            df = pd.json_normalize(results)
        else:
            df = pd.DataFrame(results)

        return df



"""
Miscellaneous functions useful for Threat Hunting and cybersecurity data analytics
"""
from __future__ import division

from builtins import input
import getpass
import math
from jellyfish import levenshtein_distance, damerau_levenshtein_distance, hamming_distance, jaro_similarity, jaro_winkler_similarity
import sys
import platform
import multiprocessing
import re

import pandas as pd
import numpy as np
from pandas.api.types import is_list_like
from math import trunc
from scipy.stats import chisquare


__all__ = ['entropy', 'entropy_per_byte', 'promptCreds', 'edit_distance']

def entropy(string):
    '''
    Calculates the Shannon entropy of a string.

    string: A string for which to compute the entropy.
    '''
    # get probability of chars in string
    prob = [ string.count(c) / len(string) for c in dict.fromkeys(list(string)) ]
    # calculate the entropy
    entropy = - sum([ p * math.log(p) / math.log(2.0) for p in prob ])
    return entropy

def entropy_per_byte(string):
    '''
    Calculates the Shannon entropy of a string, divided by the total bytes
    in the string.  This is an attempt to normalize entropy values between
    strings of different lengths.

    string: A string for which to compute the entropy per byte
    '''
    e = entropy(string)
    return e / len(string)

def promptCreds(uprompt="Username: ", pprompt="Password: "):
    '''
    Prompt the user for login credentials for some service.  This is a
    helpful convenience when using things like Jupyter notebook, where it
    may not always be obvious how to collect input from the user. The function
    returns a (username, password) tuple.

    uprompt: A string containing the username prompt. Default is "Username: ".
    pprompt: A string containing the password prompt. Default is "Password: ".
    '''
    u = input(uprompt)
    p = getpass.getpass(pprompt)
    return (u,p)

def edit_distance(str1, str2, method="damerau-levenshtein"):
    '''
    Calculate the edit distance between 'str1' and 'str2' using any of a
    number of algorithms.

    'str1', 'str2': Input strings
    'method': The algorithm to use.

    Available algorithms:
        * levenshtein
        * damerau-levenshtein (DEFAULT)
        * hamming
        * jaro
        * jaro-winkler

    Return values:
        "levenshtein", "damerau-levenshtein" and "hamming" return integers
        "jaro" and "jaro-winkler" return floats in the range of 0.0 (completely
        different) to 1.0 (identical strings).
    '''
    algos = {
        "levenshtein":levenshtein_distance,
        "damerau-levenshtein":damerau_levenshtein_distance,
        "hamming":hamming_distance,
        "jaro":jaro_similarity,
        "jaro-winkler":jaro_winkler_similarity
    }

    if not method in list(algos.keys()):
        raise ValueError("Unsupported algorithm type: %s" % method)

    if str1 is None or str2 is None or not isinstance(str1, str) or not isinstance(str2, str):
        raise TypeError("Arguments must be strings.")

    distance_function = algos[method]

    # All the jellyfish distance functions expect unicode, which is the default
    # for Python3.  If we're running in Python2, we need to convert them.
    python_version = sys.version_info
    if python_version.major == 2:
        str1 = unicode(str1)
        str2 = unicode(str2)

    return distance_function(str1, str2)

def benfords(numbers):
    '''
    Examine the distribution of the first digits in a given corpus of numbers to see
    if they correspond to Benford's Law using a chi square test.

    Benford's Law, also known as the "first digit law" or the "law of anomalous numbers"
    states that there is a specific distribution pattern of the first digits of certain 
    groups of numbers.  See https://en.wikipedia.org/wiki/Benford%27s_law for more 
    info.

    :param numbers: The set of numbers to check against Benford's Law
    :type numbers: A list-like object (list, tuple, set, Pandas DataFrame or Series) 
                  containing floats or integers

    :Return Value:

    The function returns three values in a tuple (chi2, p, counts):

      * The 'chi2' value is a float in the range 0..1 that describes how well the observed 
        distribution of first digits matched the predictions of Benford's Law.  Lower is 
        better.  
      * The 'p' value is the probability that the computed 'chi2' is significant (i.e., it 
        tells you whether the chi2 value can be trusted).  Its range is also 0..1, but in 
        this case, higher is better.  Generally speaking, if the p-value is >= 0.95 then 
        the chi2 value is considered significant.
      * 'counts' is a Pandas series where the indices are the possible first digits 1-9 and
        the values are the observed distributions of those digits. If the observed distributions
        didn't match up with Benford's law, the counts may help you identify the anomalous values.

    '''

    def _first_digit(i: float):
        # This doesn't apply to zeros!
        if i == 0:
            return np.nan
        # Make negative numbers positive
        if i < 0:
            i = abs(i)
        # If the number is between 0 and 1, multiply by 10 until it becomes > 1
        # so the repeated divisions will work
        elif i < 1:
            while i < 1:
                i *= 10

        while i >= 10:
            i //= 10
        return trunc(i)

    _BENFORDS = [
        0.301,  # 1
        0.176,  # 2
        0.125,  # 3
        0.097,  # 4
        0.079,  # 5
        0.067,  # 6
        0.058,  # 7
        0.051,  # 8
        0.046  # 9
    ]

    if not is_list_like(numbers):
        raise TypeError(f'The argument must be a list or list-like of numbers, not type {type(numbers)}.')
    if isinstance(numbers, pd.core.series.Series):
        numbers = numbers.values

    numbers = pd.DataFrame(numbers, columns=['numbers'])
    numbers['digits'] = numbers['numbers'].apply(_first_digit).dropna()

    counts = numbers['digits'].value_counts()

    # No leading zeroes!
    if 0 in counts.index:
        counts = counts.drop(0)

    # Ensure every digit 1-9 has an count, even if it's 0
    for i in range(1, 10):
        if not i in counts:
            counts[i] = 0

    # Sort by index just to be extra sure they are all in the correct
    # order
    counts = counts.sort_index()

    # Compute the actual distribution of first digits in the input
    # as a proportion of that count to the entire number of samples
    num_samples = counts.sum()
    counts = counts.apply(lambda x: x/num_samples)

    # Compare the actual distribution to Benford's Law
    chi2, p = chisquare(counts.values, _BENFORDS)

    # Return the results of the comparison, plus the observed counts
    return chi2, p, counts

# First time initialization on import

# Set Mac OS systems to use the older "fork" method of spawning 
# procs for the multiprocessing module.  For some reason the newer
# methods don't work (EOFErrors when creating Manager() objects)
system_type = platform.system()
if system_type == "Darwin":
    multiprocessing.set_start_method('fork')

def punctuation_pattern(strings, escape_quotes=False):
    '''
    Return only the non-alphanumeric characters in the input string(s).  
    White spaces in the input will be translated into underscore characters
    in the output. 

    :param strings: The input string(s) to process.
    :type strings: A single string or a list-like object of strings (e.g. a pandas Series)

    :Return Value:
    If the input is a single string, the output will also be a single string.  Otherwise
    the output will be a pandas Series of results in the same order as the input strings.
    '''

    def _get_punct_pattern(s: str) -> str:

        res = re.sub(
            '\s',
            '_',
            re.sub(
                '[a-zA-Z0-9]',
                '',
                s
            )
        )

        if escape_quotes:
            res = re.sub(
                '([\'\"])',
                r'\\\1',
                res
            )

        return res 

    if isinstance(strings, str):
        strings = [strings]
    elif not is_list_like(strings):
        raise TypeError(f'The argument must be a string or list-like of strings, not type {type(strings)}.')

    strings = pd.DataFrame(strings, columns=['strings'])
    strings['punct'] = strings['strings'].apply(_get_punct_pattern)

    res = strings['punct']
    
    # If there's only one result, it was because we passed a single string,
    # just return a single result. Otherwise return all results.
    return res[0] if res.shape[0] == 1 else res 




#!/usr/bin/env python3

import huntlib.util as util
from huntlib.decorators import future_warning
import warnings 

__all__ = ['entropy', 'entropy_per_byte', 'promptCreds', 'edit_distance']

@future_warning("The huntlib.entropy() function has been moved to huntlib.util.entropy(). Please update your code. This compatibility will be removed in a future release.")
def entropy(*args, **kwargs):
    return util.entropy(*args, **kwargs)


@future_warning("The huntlib.entropy_per_byte() function has been moved to huntlib.util.entropy_per_byte(). Please update your code. This compatibility will be removed in a future release.")
def entropy_per_byte(*args, **kwargs):
    return util.entropy_per_byte(*args, **kwargs)


@future_warning("The huntlib.promptCreds() function has been moved to huntlib.util.promptCreds(). Please update your code. This compatibility will be removed in a future release.")
def promptCreds(*args, **kwargs):
    return util.promptCreds(*args, **kwargs)


@future_warning("The huntlib.edit_distance() function has been moved to huntlib.util.edit_distance(). Please update your code. This compatibility will be removed in a future release.")
def edit_distance(*args, **kwargs):
    return util.edit_distance(*args, **kwargs)





from __future__ import print_function

import multiprocessing
import multiprocessing.managers
import platform
import time
from builtins import object
from datetime import datetime
from multiprocessing import (Manager, Process, Queue, cpu_count,
                             set_start_method)
from sys import stderr

import pandas as pd
import splunklib.client as client
import splunklib.results as results
import splunklib.binding

from huntlib.exceptions import AuthenticationErrorSearchException


class SplunkDF(object):
    '''
    The SplunkDF() class searches Splunk and returns results as a Pandas
    DataFrame.  This makes it easier to work with the search results with
    standard data analysis techniques.

    Example Usage:

        # Establish an connection to the Splunk server. Whether this is SSL/TLS
        # or not depends on the server, and you don't really get a say.
        s = SplunkDF(host=splunk_server, username="myuser", password="mypass")

        # Fetch all search results across all time
        df = s.search(spl="search index=win_events EventCode=4688")

        # Fetch only specific fields, still across all time
        df = s.search(spl="search index=win_events EventCode=4688 | table ComputerName _time New_Process_Name Account_Name Creator_Process_ID New_Process_ID Process_Command_Line")

        # Time bounded search, 2 days prior to now
        df = s.search(spl="search index=win_events EventCode=4688", days=2)

        # Time bounded search using Python datetime() values
        df = s.search(
                        spl="search index=win_events EventCode=4688",
                        start_time=datetime.now() - timedelta(days=2),
                        end_time=datetime.now()
        )

        # Time bounded search using Splunk notation
        df = s.search(
                        spl="search index=win_events EventCode=4688",
                        start_time="-2d@d",
                        end_time="@d"
        )

    SplunkDF() will raise AuthenticationErrorSearchException() during
    initialization in the event the server denied the supplied credentials.          
    '''

    splunk_conn = None # The connection to the Splunk server (Splunk client)

    def __init__(self, host=None, username=None, password=None, token=None, port=8089):
        '''
        Create the SplunkDF object and login to the Splunk server.
        '''
        try:
            if token:
                self.splunk_conn = client.connect(
                                                host=host, port=port,
                                                token=token,
                                                autoLogin=True, max_count=0,
                                                max_time=0
                )
            else:
                self.splunk_conn = client.connect(
                                                host=host, username=username,
                                                password=password, port=port,
                                                autoLogin=True, max_count=0,
                                                max_time=0
                )
        except splunklib.binding.AuthenticationError:
            raise AuthenticationErrorSearchException("Login failed.")

    def _retrieve_parallel_worker(self, job, offset_queue, page_size, search_results):

        while not offset_queue.empty():
            offset = offset_queue.get()

            paginate_args = dict(
                count=page_size,
                offset=offset
            )

            page_results = job.results(**paginate_args)

            for result in results.ResultsReader(page_results):
                if isinstance(result, dict):
                    search_results.append(result)


    def _retrieve_parallel(self, job, page_size=1000, processes=4):


        manager = Manager()
        search_results = manager.list()
        offset_queue = Queue()


        result_count = int(job['resultCount'])

        offset = 0

        while (offset < result_count):
            offset_queue.put(offset)
            offset += page_size

        workers = list()

        for _ in range(processes):
            p = Process(
                target=self._retrieve_parallel_worker,
                args=(
                    job,
                    offset_queue,
                    page_size,
                    search_results
                )
            )

            workers.append(p)
            p.start()

        for p in workers:
            p.join()

        return list(search_results)

    def _process_result(self, result, **kwargs):
        if isinstance(result, results.Message):
            if kwargs['verbose']:
                print(f"Message: {result}")
            return None

        if isinstance(result, dict):
            # Remove internal fields if requested
            if kwargs['internal_fields'] is False:
                for field in [key for key in result.keys() if key.startswith('_')]:
                    result.pop(field)
            elif isinstance(kwargs['internal_fields'], str):
                for field in list(map(lambda x: x.strip(), kwargs['internal_fields'].split(','))):
                    result.pop(field)

        return result


    def search(self, *args, **kwargs):
        '''
        Search Splunk and return the results as a list of dicts.

        spl: A string containing the Splunk search in SPL form
        mode: A string specifying the type of Splunk search to run ("normal" or "realtime")
        search_args: A dict containing any additional search parameters to pass to
              the Splunk server.
        days: Search the past X days. If provided, this supercedes both start_time
              and end_time.
        start_time: A datetime() object representing the start of the search
                    window, or a string in Splunk syntax (e.g., "-2d@d"). If used
                    without end_time, the end of the search window is the current time.
        end_time: A datetime() object representing the end of the search window, or a
                  string in Splunk syntax (e.g., "-2d@d"). If used without start_time,
                  the search start will be the earliest timestamp in Splunk.
        verbose: If True, any errors, warnings or other messages encountered
                 by the search process will be printed to stdout.  The default is False
                 (suppress these messages).
        limit: An integer describing the max number of search results to return.
        fields: A comma-separated string listing all of the fields to be returned in
                the results. If not 'None', this is appended to the end of the 'spl'
                query, like so: "| fields field1,field2,field3".  The default is '*',
                meaning all fields.
        internal_fields: Control whether or not to return Splunk's internal fields.
                If set to False, all fields with names beginning with an underscore 
                will be removed from the results.  If set to True, nothing will be removed.
                If this is a string, treat it as a comma-separated list of fields to remove
                from the results. Default is False.
        '''

        # Valid argument checking
        valid_args = [
            'spl', 
            'mode', 
            'search_args', 
            'verbose', 
            'days', 
            'start_time',
            'end_time', 
            'limit',
            'fields',
            'internal_fields',
            'processes',
            'page_size'
        ]
        for arg in kwargs.keys():
            if not arg in valid_args:
                raise TypeError(f"search() got an unexpected keyword argument '{arg}'")

        # Make sure args contain what we think they should

        # Set default values, but only if they don't already exist
        kwargs['mode'] = kwargs.get('mode', 'normal')
        kwargs['search_args'] = kwargs.get('search_args', None)
        kwargs['verbose'] = kwargs.get('verbose', False)
        kwargs['days'] = kwargs.get('days', None)
        kwargs['start_time'] = kwargs.get('start_time', None)
        kwargs['end_time'] = kwargs.get('end_time', None)
        kwargs['limit'] = kwargs.get('limit', None)
        kwargs['fields'] = kwargs.get('fields', '*')
        kwargs['internal_fields'] = kwargs.get('internal_fields', False)
        kwargs['processes'] = kwargs.get('processes', 1)
        kwargs['page_size'] = kwargs.get('page_size', 1000)

        # SPL query can be either the first positional argument or the 'spl' keyword arg
        spl = None
        if len(args) > 0:
            spl = args[0]
        elif "spl" in kwargs:
            spl = kwargs['spl']

        if spl is None:
            raise ValueError("You must specify an SPL search string.")

        # If present, internal_fields must be either bool or str
        if not (isinstance(kwargs['internal_fields'], bool) or isinstance(kwargs['internal_fields'], str)):
            raise ValueError(f"internal_fields must be a boolean or a string, not {type(kwargs['internal_fields'])}.")

        spl += f"| fields {kwargs['fields']}"

        if not kwargs['search_args'] or not isinstance(kwargs['search_args'], dict):
            kwargs['search_args'] = dict()
        kwargs['search_args']["search_mode"] = kwargs['mode']

        if kwargs['days']:
            # Search from current time backwards
            kwargs['search_args']["earliest_time"] = "-%dd" % kwargs['days']
        else:
            if kwargs['start_time']:
                # convert to string if it's a datetime
                if isinstance(kwargs['start_time'], datetime):
                    kwargs['start_time'] = kwargs['start_time'].isoformat()
                kwargs['search_args']["earliest_time"] = kwargs['start_time']
            if kwargs['end_time']:
                # convert to string if it's a datetime
                if isinstance(kwargs['end_time'], datetime):
                    kwargs['end_time'] = kwargs['end_time'].isoformat()
                kwargs['search_args']["latest_time"] = kwargs['end_time']

        if kwargs['limit']:

            kwargs['search_args']['exec_mode'] = 'blocking'
            kwargs['search_args']['search_mode'] = 'normal'
            kwargs['search_args']['max_count'] = kwargs['limit']

            job = self.splunk_conn.jobs.create(
                spl,
                **kwargs['search_args']
            )

            for res in self._retrieve_parallel(job):
                res = self._process_result(res, **kwargs)
                if res is not None:      
                    yield res
        else:
            search_results = self.splunk_conn.jobs.export(
                spl, 
                **kwargs['search_args']
            )

            reader = results.ResultsReader(search_results)

            for res in reader:
                    res = self._process_result(res, **kwargs)
                    if res is not None:
                        yield res

#    def search_df(self, spl, mode="normal", search_args=None, verbose=False,
#                  days=None, start_time=None, end_time=None, normalize=True,
#                  limit=None, fields="*", internal_fields=False, processes=1,
#                  page_size=1000):

    def search_df(self, *args, **kwargs):
        '''
        Search Splunk and return the results as a Pandas DataFrame.

        Accepts all the same arguments as the search() function
        '''

        normalize = kwargs.get('normalize', True)

        results = list()
        for hit in self.search(*args, **kwargs):
            results.append(hit)

        if normalize:
            df = pd.json_normalize(results)
        else:
            df = pd.DataFrame(results)

        return df




#!/usr/bin/env python

from functools import wraps
import time
import warnings

def retry(retries=3, delay=1):
    '''
    In the case of transient exceptions, transparently retry the wrapped 
    function a fewtimes.  If the exception persists, re-raise the exception
    to the calling process.

    :param retries: The number of retries (DEFAULT: 3)
    :type retries: integer
    :param delay: The number of seconds to sleep between retries
    :type delay: integer
    '''

    def _retry(func):
        @wraps(func) 
        def wrapper(*args, **kwargs):
            # Do the stuff 
            retval = None
            for i in range(1, retries+1):
                try:
                    retval = func(*args, **kwargs)
                except Exception as e:
                    # If we're all out of retries, raise the exception.
                    # Otherwise sleep and try again
                    if i == retries:
                        raise
                    else:
                        time.sleep(delay)
                        continue
                break
            return retval
        return wrapper

    return _retry 

def future_warning(msg=None):
    '''
    Mark a function as deprecated (warn the user, but then call
    the function anyway).
    '''

    def _deprecated(func):
        @wraps(func)

        def wrapper(*args, **kwargs):
            warnings.warn(msg, category=FutureWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return _deprecated



#!/usr/bin/env python

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='huntlib',
      version='0.5.3',
      description='A Python library to help with some common threat hunting data analysis operations',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/target/huntlib',
      author='David J. Bianco',
      author_email='david.bianco@target.com',
      packages=['huntlib'],
      tests_require=['pytest'],
      setup_requires=['pytest-runner'],
      install_requires=[
        'future',
        'splunk-sdk',
        'elasticsearch-dsl',
        'pandas',
        'numpy',
        'jellyfish>=0.8.2',
        'domaintools_api',
        'tqdm',
        'scipy'
      ],
      zip_safe=True,
      classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License"
      ]
)



__all__ = [
    'BaseSearchException',
    'InvalidRequestSearchException',
    'AuthenticationErrorSearchException',
    'UnknownSearchException'
]

class BaseSearchException(Exception):
    '''
    A generic base class for exceptions from the ElasticDF and SplunkDF classes.
    '''
    pass

class InvalidRequestSearchException(BaseSearchException):
    '''
    Used when the user's request search is syntactically correct but
    invalid (e.g., they requested more results be returned than the server
    is capable of returning).
    '''
    pass

class AuthenticationErrorSearchException(BaseSearchException):
    '''
    Raised when the search server rejected the supplied credentials.
    '''
    pass

class UnknownSearchException(BaseSearchException):
    '''
    Returned when we must raise an exception but we don't have a better
    defined exception.
    '''



import pandas as pd 
import numpy as np 

from glob import glob 

import os

__all__ = ['read_json', 'read_csv', 'flatten']

def _read_multi(read_function=None, path_or_buf=None, post_function=None, *args, **kwargs):
    """
    Given a wildcard filename pattern (which may be just a single static
    filename), expand the wildcard and read all the files into a single
    pandas DataFrame() object.  

    :param read_function: Reference to the function which will read an individual data file (e.g., pd.read_csv)
    :param path_or_buf: A wildcard specifying which file(s) to read
    :type read_function: A reference to a valid function which returns a pd.DataFrame() object
    :type path_or_buf: A `str`, `bytes` or os.PathLike object
    """

    # Make sure we have specified a read function.  This should never
    # be called by an end user, so our code should always include one,
    # but you never know.
    if not read_function:
        raise ValueError("Must specify a read function in the `read_function` arg.")

    # Make sure we have a valid type of data for `path_or_buf` in glob(),
    # otherwise raise the same exception the original pandas function 
    # would
    if not type(path_or_buf) in [str, bytes, os.PathLike]:
        raise ValueError(f"Invalid file path or buffer object type: {type(path_or_buf)}")

    combined_df = pd.DataFrame() 
    for f in glob(path_or_buf):
        temp_df = read_function(f, *args, **kwargs)
        if post_function:
            temp_df = post_function(temp_df, f)
        combined_df = combined_df.append(temp_df, ignore_index=True)

    return combined_df

def read_json(path_or_buf=None, *args, **kwargs):
    """
    A convenience wrapper for _read_multi() that uses pd.read_json as the
    read function.
    """

    return _read_multi(
        read_function=pd.read_json,
        path_or_buf=path_or_buf,
        *args,
        **kwargs
    )

def read_csv(path_or_buf=None, *args, **kwargs):
    """
    A convenience wrapper for _read_multi() that uses pd.read_csv as the
    read function.
    """

    return _read_multi(
        read_function=pd.read_csv,
        path_or_buf=path_or_buf,
        *args,
        **kwargs
    )


def flatten(obj, sep='.'):
    '''
    Given a dictionary or list that may contain other dictionaries
    or lists, recursively flatten all the values into a single 
    level and return them as a list.

    :param obj: The possibly-multilevel object to flatten
    :param type: dict or list object
    :param sep: The character to use to separate the path names. (DEFAULT '.')
    :param sep: string

    :Return Value:
    A dict containing all the entries of member dicts or lists in a single
    flat keyspace. 

    For example:

    >>> flatten({"key1": "val1", "subkeys": {"subkey1": "subval1"}})
    {'key1': 'val1', 'subkeys.subkey1': 'subval1'}

    Lists are flattened according to their indices:

    >>> flatten([1, 2, 3])
    {'0': 1, '1': 2, '2': 3}

    A more complex example:

    >>> huntlib.flatten([dict(a='a', b='b'), dict(a='a1', c='c')])
    {'0.a': 'a', '0.b': 'b', '1.a': 'a1', '1.c': 'c'}

    '''
    def _flatten(obj, keypath='', sep='.'):

        if isinstance(obj, dict):
            keypath = keypath + sep if keypath else keypath
            for k in obj:
                yield from _flatten(obj[k], keypath + str(k), sep=sep)
        elif isinstance(obj, list):
            keypath = keypath + sep if keypath else keypath
            for k in range(len(obj)):
                yield from _flatten(obj[k], keypath + str(k), sep=sep)
        else:
            yield keypath, obj

    return dict(_flatten(obj=obj, sep=sep))

def chunk(sequence, size=10):
    '''
    Given a sequence-like object, divide into chunks of `size`
    and return those as a generator. If the length of the sequence
    is not evenly divisible by the size, the final chunk will 
    contain however many items remain.

    :param sequence: The sequence to chunk up
    :param size: The chunk size 
    '''

    for chunk in [sequence[i:i+size] for i in range(0, len(sequence), size)]:
        yield chunk

